import locale
from typing import NamedTuple, Tuple

import click
from requests import Session
from tabulate import tabulate

from boiler import cli

locale.setlocale(locale.LC_ALL, '')


class CostStatistics(NamedTuple):
    count: int  # type: ignore
    total: float
    p50: float
    p90: float
    p95: float
    p99: float
    average: float


def mturk_costs(session: Session) -> Tuple[CostStatistics, CostStatistics]:
    resp = session.get('mturk/cost')
    obj = resp.json()
    video = CostStatistics(**obj['video'])
    activity = CostStatistics(**obj['activity'])
    return video, activity


def mturk_balance(session: Session) -> float:
    resp = session.get('mturk/balance')
    if not resp.ok:
        click.secho('Could not get mechanical turk balance.', fg='red')
        return 0
    return resp.json()


@click.group(name='mturk', short_help='Get information about HITs on mechanical turk')
def mturk():
    pass


@mturk.command(help='Get the available balance in our account')
@click.pass_obj
def balance(ctx):
    b = mturk_balance(ctx['session'])
    click.echo(locale.format('%.2f', b, grouping=True))


@mturk.command(help='Get historical costs from previous HITs')
@click.pass_obj
def costs(ctx):
    v, a = mturk_costs(ctx['session'])
    click.echo(locale.format_string('Number of videos: %i', v.count))
    click.echo('\nCosts per video\n' + '-' * 15)
    table = [
        ('Average:', '$' + locale.format('%.2f', v.average, grouping=True)),
        ('Median:', '$' + locale.format('%.2f', v.p50, grouping=True)),
        ('95th percentile:', '$' + locale.format('%.2f', v.p95, grouping=True)),
        ('Total:', '$' + locale.format('%.2f', v.total, grouping=True)),
    ]
    click.echo(tabulate(table, tablefmt='plain', colalign=('left', 'right')))
    click.echo('')
    click.echo(locale.format_string('Number of activities: %i', a.count))
    click.echo('\nCosts per activity\n' + '-' * 18)
    table = [
        ('Average:', '$' + locale.format('%.2f', a.average, grouping=True)),
        ('Median:', '$' + locale.format('%.2f', a.p50, grouping=True)),
        ('95th percentile:', '$' + locale.format('%.2f', a.p95, grouping=True)),
        ('Total:', '$' + locale.format('%.2f', a.total, grouping=True)),
    ]
    click.echo(tabulate(table, tablefmt='plain', colalign=('left', 'right')))


cli.add_command(mturk)



import os

from boiler import BoilerError

supported_extensions = ['mp4', 'avi']


def multipart_upload(
    client,
    video: dict,
    video_id: str,
    source_path: str,
    bucket: str,
    destination_base_directory: str = 'original',
):
    """
    Uploads a stumpf-known video to s3 for transcoding.
    """

    _, extension = os.path.splitext(source_path)
    if extension not in supported_extensions:
        raise BoilerError(f'extension={extension} unsupported for file={source_path}')
    destination_base_directory = destination_base_directory.lstrip('/')
    destination = '/'.join(
        [
            destination_base_directory,
            video['date'],
            '_'.join([video['start_time'], video['end_time']]),
            '_'.join([video['gtag'], video['location']]),
            video_id + extension,
        ]
    )
    with open(source_path, 'rb') as f:
        return client.put_object(ACL='private', Bucket=bucket, Key=destination, Body=f,)



import hashlib
from itertools import groupby
import json
from json.decoder import JSONDecodeError
from typing import BinaryIO, List, TextIO

import click
from requests import Response

from boiler.definitions import ActivityType
from boiler.models import ActivityList


def handle_request_error(r: Response) -> dict:
    if not r.ok:
        # TODO: inspect content-type to infer this
        error_text = r.text
        try:
            error_text = r.json()
        except ValueError:
            pass

        headers = dict(r.request.headers)
        headers.pop('X-Stumpf-Token', None)
        body = str(r.request.body)
        try:
            body = json.loads(body)
        except JSONDecodeError:
            pass

        return {
            'context': {
                'url': r.url,
                'method': r.request.method,
                'status': r.status_code,
                'body': body,
                'headers': headers,
            },
            'error': error_text,
        }
    return {'response': r.json()}


def exit_with(out: dict):
    click.echo(json.dumps(out, indent=2, sort_keys=True))
    if out.get('error'):
        exit(1)
    exit(0)


def sha1(file: BinaryIO) -> str:
    file.seek(0, 0)  # reset reader position
    m = hashlib.sha1()
    m.update(file.read())
    sha1 = m.hexdigest()
    file.seek(0, 0)
    return sha1


def activity_types_from_file(file: TextIO) -> List[str]:
    activity_types: List[str] = []
    for a in file.read().splitlines():
        if len(a) > 0:
            activity_types.append(ActivityType(a).value)
    return activity_types


def summarize_activities(activity_list: ActivityList):
    def activity_keyfunc(v):
        return v.activity_type

    def actor_keyfunc(v):
        return v.actor_type.value

    activity_counts = []
    for key1, group1 in activity_list.activity_types.items():
        group1_list = list(group1)
        activity_counts.append(
            {
                'name': key1.value,
                'count': len(group1_list),
                'frame_length_sum': sum([a.end - a.begin for a in group1_list]),
            }
        )

    actors = []
    for act in activity_list.activity_map.values():
        for actor in act.actors:
            actors.append(actor)
    actor_counts = []
    for key2, group2 in groupby(sorted(actors, key=actor_keyfunc), key=actor_keyfunc):
        group2_list = list(group2)
        actor_counts.append(
            {
                'name': key2,
                'count': len(group2_list),
                'frame_length_sum': sum([a.end - a.begin for a in group2_list]),
            }
        )

    return {
        'activities': {'count': len(activity_list.activity_map), 'by_type': activity_counts,},
        'actors': {'count': len(actors), 'by_type': actor_counts},
    }



"""
KW18 is a csv-type format consisting of 3 files.
A row schema for each file is defined below.
[?] indicates an un-used row field.

file.txt: contains activity information.
[?] [1:ACTIVITY_TYPE_CODE] [2:ACTIVITY_ID] [?] [4:BEGIN]
[?] [6:END] [?] [?] [?] [?] [?] [12:N_PARTICIPANTS]
[ACTOR1_ID] [ACTOR2_ID] ...
[?] [ACTOR1_BEIGN] [?] [ACTOR1_END] [?] [ACTOR2_BEIGN] [?] [ACTOR2_END] ...

file.kw18: contains geometry information
[0:actor_id] [1:actor_track_length] [2:frame]
[?] [?] [?] [?] [?] [?] [9:top_left_x] [10:top_left_y] [11:bottom_right_x] [12:bottom_right_y]

file.kw18.types: contains actor type information
[0:ACTOR_ID] [1:ACTOR_TYPE]

file.kw18.regions: contains keyframe information
[0:actor_id] [1:frame] [2:is_keyframe] [?] ...
"""
from collections import defaultdict
import csv
import enum
from pathlib import Path
import re
from typing import Any, Dict, IO, Iterator, List, NamedTuple, Optional, Set

from boiler import BoilerError, BoilerSession
from boiler.definitions import ActivityType, ActorType, all_activity_codes
from boiler.models import Activity, ActivityList, Actor, Box, Detection
from boiler.serialization import error_handler, ParseError


ActivityMapType = Dict[int, Activity]
ActorMapType = Dict[int, Actor]

comment_re = re.compile('#.*$')


class ActivityRow(enum.IntEnum):
    TYPE = 1
    ID = 2
    BEGIN = 4
    END = 6
    ACTOR_COUNT = 12
    ACTOR_LIST_START = 13


class ActivityActorDetail(enum.IntEnum):
    BEGIN = 1
    END = 3


class GeomRow(enum.IntEnum):
    ACTOR_ID = 0
    ACTOR_TRACK_LENGTH = 1
    FRAME = 2
    LEFT = 9
    TOP = 10
    RIGHT = 11
    BOTTOM = 12


class RegionsRow(enum.IntEnum):
    ACTOR_ID = 0
    FRAME = 1
    KEYFRAME = 2


class TypesRow(enum.IntEnum):
    ID = 0
    TYPE = 1


class KW18FileSet(NamedTuple):
    txt: Optional[IO[str]] = None
    types: Optional[IO[str]] = None
    kw18: Optional[IO[str]] = None
    regions: Optional[IO[str]] = None

    def read(self) -> ActivityList:
        activities: List[Activity] = []
        if self.txt is not None:
            detection_map = self.read_detections()
            actor_type_map = self.read_actor_types()
            for row in self._read_file(self.txt):
                activities.append(self._parse_activity_row(row, actor_type_map, detection_map))

        return ActivityList.create_from_activity_list(activities)

    def _upload_file(self, file: Optional[IO[str]], session: BoilerSession) -> Optional[str]:
        if file is None:
            return None
        file.seek(0)
        r = session.post('upload', files={'file': file})
        if r.status_code not in (409, 201):
            raise Exception(f'Could not upload {file.name}')
        return r.json()['id']

    def upload(self, session: BoilerSession) -> Dict[str, Any]:
        """Upload a file set as a KW18FileUpload resource."""
        data = {
            'kw18_id': self._upload_file(self.kw18, session),
            'txt_id': self._upload_file(self.txt, session),
            'types_id': self._upload_file(self.types, session),
            'regions_id': self._upload_file(self.regions, session),
        }
        r = session.get('kw18-file-upload', json=data)
        if r.status_code == 200:
            values = r.json()
            if values:
                return values[0]

        r = session.post('kw18-file-upload', json=data)
        if r.status_code != 201:
            raise Exception('Could not upload KW18')
        return r.json()

    @classmethod
    def write(cls, base_name: str, activities: ActivityList):
        serialize_to_files(base_name, activities)

    @classmethod
    def create_from_path(cls, path: Path) -> 'KW18FileSet':
        """Create a KW18FileSet from a path on disk.

        This assumes there is *exactly* one *.kw18 file in the directory and
        that the other files have the same base name.
        """
        candidates = list(path.glob('*.kw18'))
        if len(candidates) > 1:
            raise FileNotFoundError('Provided path contains more than one kw18 file')
        elif not candidates:
            raise FileNotFoundError('Provided path does not contain any kw18 files')

        prefix = candidates[0].stem
        kw18 = candidates[0].open('r')
        return cls(
            kw18=kw18,
            txt=cls._maybe_open_file(path / (prefix + '.txt')),
            types=cls._maybe_open_file(path / (prefix + '.kw18.types')),
            regions=cls._maybe_open_file(path / (prefix + '.kw18.regions')),
        )

    @classmethod
    def _maybe_open_file(cls, path: Path) -> Optional[IO[str]]:
        if path.is_file():
            return path.open('r')

        return None

    @classmethod
    @error_handler('Could not parse activity from the *.txt file')
    def _parse_activity_row(
        cls,
        row: List[str],
        actor_type_map: Dict[int, ActorType],
        detection_map: Dict[int, List[Detection]],
    ) -> Activity:
        type_id = int(row[ActivityRow.TYPE])
        if type_id not in all_activity_codes:
            raise ParseError('Unknown activity type id "{type_id}"')

        activity_type = all_activity_codes[type_id]
        activity_id = int(row[ActivityRow.ID])
        begin = int(row[ActivityRow.BEGIN])
        end = int(row[ActivityRow.END])
        actor_count = int(row[ActivityRow.ACTOR_COUNT])

        actors = []
        for actor_index in range(actor_count):
            actor_id = int(row[ActivityRow.ACTOR_LIST_START + actor_index])
            start_index = ActivityRow.ACTOR_LIST_START + actor_count + actor_index * 4

            if actor_count == 1:
                actor_begin = begin
                actor_end = end
            else:
                actor_begin = int(row[start_index + ActivityActorDetail.BEGIN])
                actor_end = int(row[start_index + ActivityActorDetail.END])

            detections = [d for d in detection_map[actor_id]]
            if actor_id not in actor_type_map:
                raise ParseError(f'Unknown actor_id={actor_id} in activity file')

            actors.append(
                Actor(
                    actor_id=actor_id,
                    actor_type=actor_type_map[actor_id],
                    begin=actor_begin,
                    end=actor_end,
                    detections=detections,
                )
            )

        return Activity(
            activity_id=activity_id,
            activity_type=activity_type,
            begin=begin,
            end=end,
            actors=actors,
        )

    @classmethod
    def _get_file_lines(cls, f: IO[str]) -> Iterator[str]:
        for line in f:
            line = comment_re.sub('', line).strip()
            if line:
                yield line

    @classmethod
    def _read_file(cls, f: IO[str]) -> Iterator[List[str]]:
        for row in csv.reader(cls._get_file_lines(f), delimiter=' '):
            yield row

    @error_handler('Could not parse actors from the *.types file')
    def read_actor_types(self) -> Dict[int, ActorType]:
        if self.types is None:
            return {}
        rows = self._read_file(self.types)
        return {int(row[TypesRow.ID]): ActorType.from_string(row[TypesRow.TYPE]) for row in rows}

    @error_handler('Could not parse interpolated frames from the *.regions file')
    def read_interpolated_frames(self) -> Dict[int, Set[int]]:
        interpolated_frames: Dict[int, Set[int]] = defaultdict(set)

        if self.regions is not None:
            for row in self._read_file(self.regions):
                if row[RegionsRow.KEYFRAME] == '0':
                    actor_id = int(row[RegionsRow.ACTOR_ID])
                    frame = int(row[RegionsRow.FRAME])
                    interpolated_frames[actor_id].add(frame)

        return interpolated_frames

    @error_handler('Could not parse detections from the *.kw18 file')
    def read_detections(self) -> Dict[int, List[Detection]]:
        detections: Dict[int, List[Detection]] = defaultdict(list)
        if self.kw18 is None:
            return detections

        interpolated_frames = self.read_interpolated_frames()
        for row in self._read_file(self.kw18):
            frame = int(row[GeomRow.FRAME])
            actor_id = int(row[GeomRow.ACTOR_ID])
            if frame not in interpolated_frames[actor_id]:
                detections[actor_id].append(
                    Detection(
                        frame=frame,
                        box=Box(
                            left=int(row[GeomRow.LEFT]),
                            right=int(row[GeomRow.RIGHT]),
                            top=int(row[GeomRow.TOP]),
                            bottom=int(row[GeomRow.BOTTOM]),
                        ),
                        keyframe=True,
                    )
                )
        return detections


def find_activity_code(activity_type: ActivityType):
    for code, activity in all_activity_codes.items():
        if activity == activity_type:
            return code
    raise BoilerError(f'no numerical code exists for {activity_type.value}')


def serialize_types(actor_map: Dict[int, Actor]):
    for actor_id, actor in actor_map.items():
        row = [actor_id, actor.actor_type.value]
        yield ' '.join([str(d) for d in row]) + '\n'


def serialize_activities(
    activity_map: Dict[int, Activity], actor_map: Dict[int, Actor], filler_char
):
    inverse_actor_map = {actor: actor_id for actor_id, actor in actor_map.items()}
    for activity_id, activity in activity_map.items():
        row = [
            filler_char,
            find_activity_code(activity.activity_type),
            activity_id,
            filler_char,
            activity.begin,
            filler_char,
            activity.end,
            filler_char,
            filler_char,
            filler_char,
            filler_char,
            filler_char,
            len(activity.actors),
        ]
        for actor in activity.actors:
            actor_id = inverse_actor_map[actor]
            row.append(str(actor_id))

        if len(activity.actors) > 1:
            for actor in activity.actors:
                row += [filler_char, actor.begin, filler_char, actor.end]

        yield ' '.join([str(d) for d in row]) + '\n'


def serialize_geom(actor_map: Dict[int, Actor], filler_char):
    for actor_id, actor in actor_map.items():
        for detection in actor.detections:
            if detection.keyframe:
                row = [
                    actor_id,
                    actor.end - actor.begin,
                    detection.frame,
                    filler_char,
                    filler_char,
                    filler_char,
                    filler_char,
                    filler_char,
                    filler_char,
                ]
                row += list(detection.box)
                yield ' '.join([str(d) for d in row]) + '\n'


def serialize_to_files(
    output_basename: str, activity_list: ActivityList, filler_char='0',
):
    actor_map = activity_list.actor_map
    activity_map = activity_list.activity_map

    with open(output_basename + '.txt', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_activities(activity_map, actor_map, filler_char):
            yamlfile.write(line)
    with open(output_basename + '.kw18.types', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_types(actor_map):
            yamlfile.write(line)
    with open(output_basename + '.kw18', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_geom(actor_map, filler_char):
            yamlfile.write(line)



from math import ceil
import sys
from typing import Tuple

import click
import gitlab
from gitlab.exceptions import GitlabAuthenticationError
from gitlab.v4.objects import Issue

from boiler import BOILER_CONFIG_FILE, cli, get_config_value


ISSUE_TEMPLATE = """
AUDITOR:

DATE AUDIT COMPLETE: MM/DD/YY

* Number of Janitor Tasks: {tasks}
* Number of Janitor Tasks to 95% complete: {to_95}

Project to charge (with KWIK ID): IARPA DIVA-Option 2-Annotation Generation K002093-00-T27

Path to janitor task: <http://admin.stumpf.avidannotations.com/#/tasks?video={video_name}>


**Notes:** You will need to correct or approve the annotations for the janitor activities
associated with this clip. You should *not* review the entire clip.
"""  # noqa

TITLE_TEMPLATE = 'DIVA M2 Janitor {video_name}'  # noqa


@click.group(name='janitor')
@click.pass_obj
def janitor(ctx):
    try:
        ctx['kwgitlab'] = gitlab.Gitlab(
            ctx['gitlab_url'],
            timeout=10,
            private_token=get_config_value(BOILER_CONFIG_FILE, 'kwgitlab_token'),
        )
        ctx['kwgitlab'].auth()
        ctx['kwgitlab_project'] = ctx['kwgitlab'].projects.get(ctx['gitlab_project_id'])
    except GitlabAuthenticationError:
        click.echo(
            click.style(
                "You are attempting to perform an authorized operation but you aren't authenticated with kwgitlab.\n"  # noqa
                'Run the following command: boiler login kwgitlab',
                fg='yellow',
            ),
            err=True,
        )
        sys.exit(1)


def _create_or_reopen_issue(ctx, video_name, tasks, activities, **kwargs) -> Tuple[Issue, bool]:
    # TODO: how to handle if there's more than one issue
    existing_issue = ctx['kwgitlab_project'].search(
        'issues', TITLE_TEMPLATE.format(video_name=video_name)
    )

    if existing_issue:
        existing_issue = ctx['kwgitlab_project'].issues.get(existing_issue[0]['iid'])

        if existing_issue.state == 'closed':
            existing_issue.state_event = 'reopen'
            existing_issue.save()
        return existing_issue, False

    to_95 = max(ceil(tasks - 0.05 * activities), 0)
    labels = ['DIVA', 'Janitor', 'Unassigned']
    if to_95 > 0:
        labels.append('High')
    else:
        labels.append('Medium')

    issue = ctx['kwgitlab_project'].issues.create(
        {
            'title': TITLE_TEMPLATE.format(video_name=video_name),
            'description': ISSUE_TEMPLATE.format(
                video_name=video_name, activities=activities, tasks=tasks, to_95=to_95,
            ),
            'labels': labels,
        },
        retry_transient_errors=True,
    )
    return issue, True


def _maybe_create_video_issue_association(ctx, video_id, issue_url):
    r = ctx['session'].post('task/issue', json={'video_id': video_id, 'issue_url': issue_url})

    # ignore conflict errors
    if not r.ok and r.status_code != 409:
        r.raise_for_status()


@janitor.command(name='create-issues')
@click.pass_obj
def create_issues(ctx):
    r = ctx['session'].get('video-pipeline/gunrunner/pending-janitor-tasks')
    for pending_task in r.json():
        issue, created = _create_or_reopen_issue(ctx, **pending_task)

        _maybe_create_video_issue_association(ctx, pending_task['video_id'], issue.web_url)

        if created:
            click.echo(f'created issue  {issue.iid:5}: {issue.web_url}')
        else:
            click.echo(f'reopened issue {issue.iid:5}: {issue.web_url}')


cli.add_command(janitor)



from collections import defaultdict
import functools
import json
import operator
from pathlib import Path
import re
import subprocess
import sys
from typing import Dict
import uuid
from uuid import UUID

import attr
import boto3
from botocore.exceptions import WaiterError
import click

from boiler import BoilerSession, cli
from boiler.commands.utils import exit_with, handle_request_error
from boiler.definitions import CameraLocation, ReleaseBatches, Scenarios

video_name_regexp = re.compile(
    r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})'
    r'\.(?P<start_hour>\d{2})-(?P<start_minute>\d{2})-(?P<start_second>\d{2})'
    r'\.(?P<end_hour>\d{2})-(?P<end_minute>\d{2})-(?P<end_second>\d{2})'
    r'\.(?P<location>[^.]+)'
    r'\.(?P<gtag>G[^.]+)$'
)
BUCKET = 'diva-mturk-ingestion-prod'

ELASTIC_TRANSCODER_PIPELINE_ID = '1536759483093-n34917'
ELASTIC_TRANSCODER_PORTRAIT_PRESET_ID = '1582841064651-nyezpa'
ELASTIC_TRANSCODER_LANDSCAPE_PRESET_ID = '1536756735287-cx0qtx'


def bold(x, **kwargs):
    return click.style(x, bold=True, **kwargs)


@attr.s(auto_attribs=True)
class VideoInfo:
    frame_rate: float
    duration: float
    width: int
    height: int
    name: str
    portrait: bool

    @classmethod
    def create(cls, file: str) -> 'VideoInfo':
        ps = subprocess.Popen(
            ['ffprobe', '-print_format', 'json', '-show_streams', file],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        output = ps.communicate()[0]
        stream = json.loads(output)['streams'][0]
        stream['framerate'] = round(
            functools.reduce(
                operator.truediv, map(int, stream.get('avg_frame_rate', '').split('/'))
            )
        )
        return cls(
            frame_rate=stream['framerate'],
            duration=float(stream['duration']),
            width=int(stream['width']),
            height=int(stream['height']),
            name=str(Path(file).stem),
            portrait=int(stream['height']) > int(stream['width']),
        )

    def s3_path(self, id: UUID) -> str:
        _m = video_name_regexp.match(self.name)
        if _m is None:
            raise Exception('Could not parse video name')
        m = _m.groupdict()
        date = f'{m["year"]}-{m["month"]}-{m["day"]}'
        time = (
            f'{m["start_hour"]}-{m["start_minute"]}-{m["start_second"]}'
            f'{m["end_hour"]}-{m["end_minute"]}-{m["end_second"]}'
        )
        return f'uuid-encoded/{date}/{time}/{m["gtag"]}_{m["location"]}/{id}.mp4'


def _transcode_video(upload_id: str, video_info: VideoInfo):
    transcoder = boto3.client('elastictranscoder', region_name='us-east-1')
    job = transcoder.create_job(
        PipelineId=ELASTIC_TRANSCODER_PIPELINE_ID,
        Input={'Key': f'uploads/original/{upload_id}', 'FrameRate': 'auto'},
        Output={
            'Key': f'uploads/encoded/{upload_id}',
            'PresetId': ELASTIC_TRANSCODER_PORTRAIT_PRESET_ID
            if video_info.portrait
            else ELASTIC_TRANSCODER_LANDSCAPE_PRESET_ID,
        },
    )

    try:
        transcoder.get_waiter('job_complete').wait(Id=job['Job']['Id'], WaiterConfig={'Delay': 10})
    except WaiterError:
        raise Exception(f'an error occurred while transcoding {upload_id}')


def ingest_video(video_path: str, release_batch: str, phase: str, session: BoilerSession):
    video_name = Path(video_path).stem
    video_info = VideoInfo.create(video_path)
    upload_id = str(uuid.uuid4())

    if Path(video_path).suffix != '.avi':
        raise Exception('Expected a non-transcoded video')

    r = session.get('video', json={'name': video_name})
    r.raise_for_status()

    if r.json():
        click.echo(f'name={video_name} already exists', err=True)
        return

    try:
        s3 = boto3.client('s3')
        click.echo(f'uploading {video_name}..')
        s3.upload_file(video_path, BUCKET, f'uploads/original/{upload_id}')
    except Exception:
        click.echo('Could not upload file to S3.  Have you set up your credentials?', err=True)
        raise

    click.echo(f'transcoding {video_name}..')
    _transcode_video(upload_id, video_info)

    # create stumpf video and move encoded upload to correct location
    r = session.post('video/name', json={'video_name': video_name, 'release_batch': release_batch})
    if not r.ok:
        raise Exception(f'Could not generate new video for {video_name}')

    src_path = f'uploads/encoded/{upload_id}'
    dest_path = f'{video_info.s3_path(r.json()["id"])}'
    s3.copy_object(Bucket=BUCKET, CopySource={'Bucket': BUCKET, 'Key': src_path}, Key=dest_path)
    s3.delete_object(Bucket=BUCKET, Key=f'uploads/original/{upload_id}')
    s3.delete_object(Bucket=BUCKET, Key=src_path)

    resp = session.patch(
        f'video/{r.json()["id"]}',
        json={
            'duration': video_info.duration,
            'frame_rate': video_info.frame_rate,
            'height': video_info.height,
            'width': video_info.width,
            'phase': phase,
            'path': f'https://s3.amazonaws.com/{BUCKET}/{dest_path}',
        },
    )
    if not resp.ok:
        raise Exception(f'Could not set video metadata on {video_path}')
    click.echo(f'ingested {video_name} -> {r.json()["id"]}.')


@click.group(name='video', short_help='ingest and query video')
@click.pass_obj
def video(ctx):
    pass


@video.command(name='status', help='status of video annotation')
@click.argument('name', type=click.STRING)
@click.pass_obj
def status(ctx, name):
    r = ctx['session'].get('video', params={'name': name})
    if not r.ok and r.status_code != 404:
        exit_with(handle_request_error(r))
    elif r.status_code == 404 or not r.json():
        click.echo('Video not found, has it been ingested?', err=True)
        sys.exit(1)

    video_id = r.json()[0]['id']
    r = ctx['session'].get(f'video/{video_id}/status').json()

    click.echo('-' * len(name))
    click.secho(name, bold=True)
    click.echo('-' * len(name))
    click.echo(f'https://admin.stumpf.avidannotations.com/#/video/{video_id}')

    status = r['status']
    fg = 'red'
    if 'complete' in status:
        fg = 'green'
    click.echo(f'status: {bold(status, fg=fg)}')
    click.echo(f"activity types: {bold(str(r['activity_type_count']))}")
    for cauldron in r['cauldrons']:
        click.echo(f"\ncauldron: {bold(cauldron['path'])}")
        click.echo(f"  * vendor: {bold(cauldron['vendor'])}")
        click.echo(f"  * set: {bold(cauldron['set_name'])}")

    total = 0
    total_counts: Dict[str, int] = defaultdict(lambda: 0)
    activities: Dict[str, Dict[str, int]] = defaultdict(lambda: {})
    for a in r['activity_types']:
        total += a['count']
        total_counts[a['activity_type']] += a['count']
        activities[a['activity_type']][a['status']] = a['count']

    click.echo(f'\n{total} activities:')
    for name in sorted(activities.keys()):
        summary = activities[name]
        fg = 'green' if summary.get('good') == total_counts[name] else 'yellow'

        click.echo(f'  * {bold(name, fg=fg)} ({total_counts[name]})')
        for status in sorted(summary.keys()):
            if status == 'good':
                continue
            count = summary[status]
            click.echo(f'    * {status} ({count})')


@video.command(name='search', help='search for video')
@click.option('--name', type=click.STRING)
@click.option('--gtag', type=click.STRING)
@click.option('--location', type=click.Choice([e.value for e in CameraLocation]))
@click.option('--release-batch', type=click.Choice([e.value for e in ReleaseBatches]))
@click.option('--scenario', type=click.Choice([e.value for e in Scenarios]))
@click.option('--phase', type=click.STRING)
@click.option('--page', type=click.INT)
@click.option('--size', type=click.INT)
@click.pass_obj
def search(ctx, **kwargs):
    data = {}
    for key, value in kwargs.items():
        if value is not None:
            data[key] = value
    r = ctx['session'].get('video', params=data)
    exit_with(handle_request_error(r))


@video.command(name='add', short_help='ingest video into stumpf from file')
@click.argument(
    'videos',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    nargs=-1,
    required=True,
)
@click.option(
    '--release-batch', type=click.Choice([e.value for e in ReleaseBatches]), required=True
)
@click.option('--phase', type=click.Choice(['m1', 'm2', 'meva', 'none']), required=True)
@click.pass_obj
def add(ctx, videos, release_batch, phase):
    """Ingest locally transcoded video files into Stumpf.

    \b
    Videos are found in the following SMB shares:
    * //diva-nas06/m1-complete   (all m1 locations)
    * //diva-nas01/m2-albany     (alb)
    * //diva-nas03/m2n-keu       (gdb, Greo1)
    * //diva-nas05/m2-unimore    (uni)

    Mount with:
    \b
    mount -t cifs -o username=<user>,password=<password>,ro <share> <local>

    This command requires a working `ffprobe` in your path.
    """
    phase = None if phase == 'none' else phase

    for video in videos:
        ingest_video(video, release_batch, phase, ctx['session'])


@video.command(name='summary', short_help='get a summary of all videos')
@click.argument('output', type=click.File('wb'), default='-')
@click.pass_obj
def get_video_summary(ctx, output):
    """
    Get a summary of all videos in csv format.  By default, this will print to
    the terminal.  Add a path argument to write to a file.
    """
    r = ctx['session'].get('video/summary')
    output.write(r.content)


cli.add_command(video)



import click

from boiler import cli
from boiler.commands.utils import activity_types_from_file, exit_with, handle_request_error
from boiler.definitions import AnnotationVendors


@click.group(name='vendor', short_help='Manage annotation vendors')
@click.pass_obj
def vendor(ctx):
    pass


@vendor.command(name='search')
@click.pass_obj
def search(ctx, **kwargs):
    r = ctx['session'].get('annotation-vendor')
    exit_with(handle_request_error(r))


@vendor.command(name='dispatch')
@click.option('--name', type=click.Choice([v.value for v in AnnotationVendors]), required=True)
@click.option('--video-name', type=click.STRING, required=True)
@click.option('--activity-type-list', type=click.File(mode='r'), required=True)
@click.option('--set-name', type=click.STRING, required=True)
@click.option('--annotation-repo-path', type=click.STRING, required=True)
@click.pass_obj
def dispatch(ctx, name, video_name, activity_type_list, set_name, annotation_repo_path):
    activity_types = activity_types_from_file(activity_type_list)
    data = {
        'video_name': video_name,
        'vendor_name': name,
        'activity_types': activity_types,
        'set_name': set_name,
        'annotation_repo_path': annotation_repo_path,
    }
    r = ctx['session'].post('video-pipeline/annotate', json=data)
    exit_with(handle_request_error(r))


cli.add_command(vendor)



"""
These models only provide the necessary properties for ingesting, exporting,
and validation of diva annotations for a SINGLE clip.  They are not intended
to map directly to stumpf models.

They are only useful internally to this library, and are intended as a
translation layer for data either before or after it exists within the stumpf
system.
"""
import logging
from typing import Any, Dict, Iterable, Iterator, List, NamedTuple, Optional, Tuple

import attr

from boiler import BoilerError
from boiler.definitions import ActivityPipelineStatuses, ActivityType, ActorType

logger = logging.getLogger(__name__)


def sort_detections(detections: List['Detection']) -> List['Detection']:
    return sorted(detections, key=lambda d: d.frame)


def get_next_keyframe(arr: List['Detection']) -> 'Detection':
    for d in arr:
        if d.keyframe:
            return d
    raise BoilerError('There are frames after the final keyframe!')


def interpolate_point(a, b, delta):
    return round(((1 - delta) * a) + (delta * b))


class Box(NamedTuple):
    left: int
    top: int
    right: int
    bottom: int

    def __and__(self, other: 'Box') -> 'Box':
        """Return intersection of two bounding boxes.

        For non-intersecting boxes, this will return a box of zero area.
        """
        left = max(self.left, other.left)
        right = min(self.right, other.right)
        top = max(self.top, other.top)
        bottom = min(self.bottom, other.bottom)
        return Box(left=left, right=max(left, right), top=top, bottom=max(top, bottom))

    @property
    def width(self) -> int:
        return self.right - self.left

    @property
    def height(self) -> int:
        return self.bottom - self.top

    @property
    def area(self) -> int:
        return self.width * self.height

    @property
    def aspect_ratio(self) -> float:
        height = self.height
        if height == 0:
            return 0
        return self.width / self.height

    @property
    def center(self) -> Tuple[float, float]:
        return (self.left + self.right) / 2, (self.top + self.bottom) / 2

    def interpolate(self, other: 'Box', distance: float) -> 'Box':
        left = interpolate_point(self.left, other.left, distance)
        top = interpolate_point(self.top, other.top, distance)
        right = interpolate_point(self.right, other.right, distance)
        bottom = interpolate_point(self.bottom, other.bottom, distance)
        return Box(left=left, top=top, right=right, bottom=bottom)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'Box':
        return cls(left=d['left'], right=d['right'], top=d['top'], bottom=d['bottom'])

    def fix_order(self) -> 'Box':
        left = self.left
        right = self.right
        top = self.top
        bottom = self.bottom

        if left > right:
            left, right = right, left
        elif left == right:
            logger.warning('Zero width bounding box encountered')
            right = left + 1

        if top > bottom:
            top, bottom = bottom, top
        elif bottom == top:
            logger.warning('Zero height bounding box encountered')
            bottom = top + 1

        return Box(left=left, right=right, top=top, bottom=bottom)


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class Detection:
    frame: int
    box: Box
    keyframe: bool = False

    @property
    def area(self) -> int:
        return self.box.area

    def interpolate(self, other: 'Detection', frame: int) -> 'Detection':
        distance = (frame - self.frame) / (other.frame - self.frame)
        box = self.box.interpolate(other.box, distance)
        return Detection(frame=frame, box=box, keyframe=False)

    def fix_order(self) -> 'Detection':
        return attr.evolve(self, box=self.box.fix_order())


@attr.s(auto_attribs=True, kw_only=True, frozen=True, hash=False)
class Actor:
    actor_type: ActorType
    begin: int
    end: int
    detections: List[Detection] = attr.ib(converter=sort_detections)
    actor_id: int

    @property
    def keyframes(self):
        return filter(lambda d: d.keyframe, self.detections)

    def pruned(self) -> 'Actor':
        return attr.evolve(self, detections=list(self.keyframes))

    # Used for deduplicating lists of actors when serializing
    def __hash__(self) -> int:
        first_detection = None
        if self.detections:
            first_detection = self.detections[0]
        return hash((self.actor_type.value, self.begin, self.end, first_detection))

    def inject_keyframe(self, frame: int) -> 'Actor':
        """Return the actor with a new keyframe linearly interpolated.

        If the frame is already a keyframe or if the frame is outside of the
        valid range [begin, end], then no change will be made.
        """
        new_detections: List[Detection] = []
        for index in range(len(self.detections) - 1):
            d1 = self.detections[index]
            d2 = self.detections[index + 1]
            new_detections.append(d1)

            if d1.frame == frame:
                return self
            elif d1.frame < frame < d2.frame:
                interpolated = attr.evolve(d1.interpolate(d2, frame), keyframe=True)
                new_detections.append(interpolated)

        new_detections.append(self.detections[-1])
        return attr.evolve(self, detections=new_detections)

    def truncate_framerange(self) -> 'Actor':
        """Return the actor with frame range truncated to the detection range."""
        begin = max(self.begin, self.detections[0].frame)
        end = min(self.end, self.detections[-1].frame)
        return attr.evolve(self, begin=begin, end=end)

    def truncate_detections(self) -> 'Actor':
        return attr.evolve(
            self, detections=[d for d in self.detections if self.begin <= d.frame <= self.end]
        )


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class Activity:
    activity_type: ActivityType
    begin: int
    end: int
    actors: List[Actor]
    activity_id: int
    status: Optional[ActivityPipelineStatuses] = None

    def pruned(self) -> 'Activity':
        actors = [a.pruned() for a in self.actors]
        return attr.evolve(self, actors=actors)

    def truncate(self) -> 'Activity':
        """Remove extra detections and truncate frame range."""
        if not self.actors:
            logger.warning('Activity does not contain any tracks')
            return self

        new_actors = [
            a.inject_keyframe(a.begin)
            .inject_keyframe(a.end)
            .truncate_framerange()
            .truncate_detections()
            for a in self.actors
        ]
        begin = self.actors[0].begin
        end = self.actors[0].end
        for actor in self.actors[1:]:
            begin = min(begin, actor.begin)
            end = max(end, actor.end)

        return attr.evolve(self, begin=begin, end=end, actors=new_actors)


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class ActivityList:
    activity_map: Dict[int, Activity] = attr.ib()
    actor_map: Dict[int, Actor] = attr.ib()

    @classmethod
    def create_from_activity_list(cls, activities: Iterable[Activity]) -> 'ActivityList':
        activity_map: Dict[int, Activity] = {}
        actor_map: Dict[int, Actor] = {}

        # TODO: The data model used for actors doesn't perfectly match with
        # either the database or annotation files.  The files themselves store
        # multiple "tracks" which can be joined by actor_id.  Instead of an
        # explicit "track_id", they use an (activity_id, actor_id) pair with
        # detections associated with that pair.  In the boiler representation,
        # there is no way to distinguish detections associated with a specific
        # "track" in this sense, so the frame range for which an actor is
        # involved in an activity is lost.  As a work around, we generate a new
        # actor_id for actors containing more than one track.  In practice,
        # this is rare.
        max_actor_id = 0

        for activity in activities:
            for actor in activity.actors:
                max_actor_id = max(actor.actor_id, max_actor_id)

        for activity in activities:
            if activity.activity_id in activity_map:
                raise Exception('Duplicate activity_ids detected')

            actors: List[Actor] = []
            for actor in activity.actors:
                if actor.actor_id in actor_map and actor != actor_map[actor.actor_id]:
                    max_actor_id += 1
                    actor_id = max_actor_id
                else:
                    actor_id = actor.actor_id

                actors.append(attr.evolve(actor, actor_id=actor_id))
                actor_map[actor_id] = actor

            activity_map[activity.activity_id] = attr.evolve(activity, actors=actors)

        return cls(activity_map=activity_map, actor_map=actor_map)

    @property
    def activity_types(self) -> Dict[ActivityType, Iterable[Activity]]:
        # TODO: make this more efficient
        activity_types = {a.activity_type for a in self.activity_map.values()}
        return {
            at: list(filter(lambda a: a.activity_type == at, self.activity_map.values()))
            for at in activity_types
        }

    def __iter__(self) -> Iterator[Activity]:
        for activity in self.activity_map.values():
            yield activity

    def __len__(self) -> int:
        return len(self.activity_map)



from pathlib import Path

import pytest

from boiler.serialization.refiner import deserialize_answer
from boiler.validate import validate_activity

HERE = Path(__file__).parent


@pytest.fixture
def refiner_answer():
    with (HERE / 'refiner_answer.json').open('r') as f:
        yield f.read()


def test_parse_answer(refiner_answer):
    activity = deserialize_answer(refiner_answer)
    validate_activity(activity)
    assert len(activity.actors) == 1



from typing import List

from factory import Factory, Faker, lazy_attribute, SubFactory

from boiler import definitions, models

MAX_FRAME = 30 * 60 * 5


class BoxFactory(Factory):
    class Meta:
        model = models.Box

    left: int = Faker('random_int', min=0, max=1070)
    top: int = Faker('random_int', min=0, max=750)

    @lazy_attribute
    def right(self) -> int:
        return Faker('random_int', min=self.left + 10, max=1080).generate()

    @lazy_attribute
    def bottom(self) -> int:
        return Faker('random_int', min=self.top + 10, max=760).generate()


class DetectionFactory(Factory):
    class Meta:
        model = models.Detection

    frame: int = Faker('random_int', min=0, max=MAX_FRAME)
    box: models.Box = SubFactory(BoxFactory)
    keyframe: bool = Faker('random_element', elements=[True, False])


class ActorFactory(Factory):
    class Meta:
        model = models.Actor

    actor_id: int = Faker('random_int', min=0, max=10000)
    actor_type: str = Faker('random_element', elements=definitions.ActorType)
    begin: int = Faker('random_int', min=0, max=MAX_FRAME - 10)

    @lazy_attribute
    def end(self) -> int:
        return Faker('random_int', min=self.begin + 10, max=MAX_FRAME).generate()

    @lazy_attribute
    def detections(self) -> List[models.Detection]:
        return [
            DetectionFactory(frame=self.begin, keyframe=True),
            DetectionFactory(frame=self.end, keyframe=True),
        ]


class ActivityFactory(Factory):
    class Meta:
        model = models.Activity

    activity_id: int = Faker('random_int', min=0, max=10000)
    activity_type: definitions.ActivityType = Faker(
        'random_element', elements=definitions.ActivityType
    )
    begin: int = Faker('random_int', min=0, max=MAX_FRAME - 10)

    @lazy_attribute
    def end(self) -> int:
        return Faker('random_int', min=self.begin + 10, max=MAX_FRAME).generate()

    @lazy_attribute
    def actors(self) -> List[models.Actor]:
        return [
            ActorFactory(),
            ActorFactory(),
        ]



import enum
from typing import Type, TypeVar


T = TypeVar('T', bound='StrEnum')


# This is needed to get mypy to validate enum's correctly.  By default,
# enum's have value type Any, which means that it is never validated.
class StrEnum(enum.Enum):
    @property
    def value(self) -> str:
        return super(StrEnum, self).value

    @classmethod
    def from_string(cls: Type[T], value: str) -> T:
        return cls(value.lower())


class ActorType(StrEnum):
    BAG = 'bag'
    BIKE = 'bicycle'
    OTHER = 'other'
    PERSON = 'person'
    QUEUE = 'queue'
    RECEPTACLE = 'receptacle'
    VEHICLE = 'vehicle'

    @classmethod
    def from_string(cls, value: str) -> 'ActorType':
        value = value.lower()
        if value == 'bike':
            value = 'bicycle'
        return cls(value)


actor_codes = {
    ActorType.BAG: 'b',
    ActorType.BIKE: 'k',
    ActorType.OTHER: 'o',
    ActorType.PERSON: 'p',
    ActorType.QUEUE: 'q',
    ActorType.RECEPTACLE: 'o',  # https://gitlab.com/diva-mturk/stumpf-diva/issues/32#note_285508683
    ActorType.VEHICLE: 'v',
}


class CurrentActivityType(StrEnum):
    HAND_INTERACTS_WITH_PERSON = 'hand_interacts_with_person'
    PERSON_ABANDONS_PACKAGE = 'person_abandons_package'
    PERSON_ASCENDS_STAIRS = 'person_ascends_stairs'
    PERSON_CARRIES_HEAVY_OBJECT = 'person_carries_heavy_object'
    PERSON_CLOSES_FACILITY_DOOR = 'person_closes_facility_door'
    PERSON_CLOSES_TRUNK = 'person_closes_trunk'
    PERSON_CLOSES_VEHICLE_DOOR = 'person_closes_vehicle_door'
    PERSON_DESCENDS_STAIRS = 'person_descends_stairs'
    PERSON_DISCARDS_TRASH = 'person_discards_trash'
    PERSON_DOFFS_JACKET = 'person_doffs_jacket'
    PERSON_DONS_JACKET = 'person_dons_jacket'
    PERSON_DRINKS = 'person_drinks'
    PERSON_EMBRACES_PERSON = 'person_embraces_person'
    PERSON_ENTERS_SCENE_THROUGH_STRUCTURE = 'person_enters_scene_through_structure'
    PERSON_ENTERS_VEHICLE = 'person_enters_vehicle'
    PERSON_EXITS_SCENE_THROUGH_STRUCTURE = 'person_exits_scene_through_structure'
    PERSON_EXITS_VEHICLE = 'person_exits_vehicle'
    PERSON_INTERACTS_WITH_LAPTOP = 'person_interacts_with_laptop'
    PERSON_LEADS_ANIMAL = 'person_leads_animal'
    PERSON_LOADS_VEHICLE = 'person_loads_vehicle'
    PERSON_OPENS_FACILITY_DOOR = 'person_opens_facility_door'
    PERSON_OPENS_TRUNK = 'person_opens_trunk'
    PERSON_OPENS_VEHICLE_DOOR = 'person_opens_vehicle_door'
    PERSON_PICKS_UP_OBJECT = 'person_picks_up_object'
    PERSON_PULLS_OBJECT = 'person_pulls_object'
    PERSON_PURCHASES = 'person_purchases'
    PERSON_PUSHES_OBJECT = 'person_pushes_object'
    PERSON_PUTS_DOWN_OBJECT = 'person_puts_down_object'
    PERSON_READS_DOCUMENT = 'person_reads_document'
    PERSON_RIDES_BICYCLE = 'person_rides_bicycle'
    PERSON_RUMMAGES_IN_PARCEL = 'person_rummages_in_parcel'
    PERSON_SITS_DOWN = 'person_sits_down'
    PERSON_STANDS_UP = 'person_stands_up'
    PERSON_STEALS_OBJECT = 'person_steals_object'
    PERSON_TALKS_ON_PHONE = 'person_talks_on_phone'
    PERSON_TALKS_TO_PERSON = 'person_talks_to_person'
    PERSON_TEXTS_ON_PHONE = 'person_texts_on_phone'
    PERSON_TRANSFERS_OBJECT = 'person_transfers_object'
    PERSON_UNLOADS_VEHICLE = 'person_unloads_vehicle'
    VEHICLE_DROPS_OFF_PERSON = 'vehicle_drops_off_person'
    VEHICLE_MAKES_U_TURN = 'vehicle_makes_u_turn'
    VEHICLE_PICKS_UP_PERSON = 'vehicle_picks_up_person'
    VEHICLE_REVERSES = 'vehicle_reverses'
    VEHICLE_STARTS = 'vehicle_starts'
    VEHICLE_STOPS = 'vehicle_stops'
    VEHICLE_TURNS_LEFT = 'vehicle_turns_left'
    VEHICLE_TURNS_RIGHT = 'vehicle_turns_right'


class ActivityType(StrEnum):
    # cannot extend the previous enum class:
    # https://docs.python.org/3/library/enum.html#restricted-enum-subclassing
    HAND_INTERACTS_WITH_PERSON = 'hand_interacts_with_person'
    PERSON_ABANDONS_PACKAGE = 'person_abandons_package'
    PERSON_ASCENDS_STAIRS = 'person_ascends_stairs'
    PERSON_CARRIES_HEAVY_OBJECT = 'person_carries_heavy_object'
    PERSON_CLOSES_FACILITY_DOOR = 'person_closes_facility_door'
    PERSON_CLOSES_TRUNK = 'person_closes_trunk'
    PERSON_CLOSES_VEHICLE_DOOR = 'person_closes_vehicle_door'
    PERSON_DESCENDS_STAIRS = 'person_descends_stairs'
    PERSON_DISCARDS_TRASH = 'person_discards_trash'
    PERSON_DOFFS_JACKET = 'person_doffs_jacket'
    PERSON_DONS_JACKET = 'person_dons_jacket'
    PERSON_DRINKS = 'person_drinks'
    PERSON_EMBRACES_PERSON = 'person_embraces_person'
    PERSON_ENTERS_SCENE_THROUGH_STRUCTURE = 'person_enters_scene_through_structure'
    PERSON_ENTERS_VEHICLE = 'person_enters_vehicle'
    PERSON_EXITS_SCENE_THROUGH_STRUCTURE = 'person_exits_scene_through_structure'
    PERSON_EXITS_VEHICLE = 'person_exits_vehicle'
    PERSON_INTERACTS_WITH_LAPTOP = 'person_interacts_with_laptop'
    PERSON_LEADS_ANIMAL = 'person_leads_animal'
    PERSON_LOADS_VEHICLE = 'person_loads_vehicle'
    PERSON_OPENS_FACILITY_DOOR = 'person_opens_facility_door'
    PERSON_OPENS_TRUNK = 'person_opens_trunk'
    PERSON_OPENS_VEHICLE_DOOR = 'person_opens_vehicle_door'
    PERSON_PICKS_UP_OBJECT = 'person_picks_up_object'
    PERSON_PULLS_OBJECT = 'person_pulls_object'
    PERSON_PURCHASES = 'person_purchases'
    PERSON_PUSHES_OBJECT = 'person_pushes_object'
    PERSON_PUTS_DOWN_OBJECT = 'person_puts_down_object'
    PERSON_READS_DOCUMENT = 'person_reads_document'
    PERSON_RIDES_BICYCLE = 'person_rides_bicycle'
    PERSON_RUMMAGES_IN_PARCEL = 'person_rummages_in_parcel'
    PERSON_SITS_DOWN = 'person_sits_down'
    PERSON_STANDS_UP = 'person_stands_up'
    PERSON_STEALS_OBJECT = 'person_steals_object'
    PERSON_TALKS_ON_PHONE = 'person_talks_on_phone'
    PERSON_TALKS_TO_PERSON = 'person_talks_to_person'
    PERSON_TEXTS_ON_PHONE = 'person_texts_on_phone'
    PERSON_TRANSFERS_OBJECT = 'person_transfers_object'
    PERSON_UNLOADS_VEHICLE = 'person_unloads_vehicle'
    VEHICLE_DROPS_OFF_PERSON = 'vehicle_drops_off_person'
    VEHICLE_MAKES_U_TURN = 'vehicle_makes_u_turn'
    VEHICLE_PICKS_UP_PERSON = 'vehicle_picks_up_person'
    VEHICLE_REVERSES = 'vehicle_reverses'
    VEHICLE_STARTS = 'vehicle_starts'
    VEHICLE_STOPS = 'vehicle_stops'
    VEHICLE_TURNS_LEFT = 'vehicle_turns_left'
    VEHICLE_TURNS_RIGHT = 'vehicle_turns_right'

    # activities used in the past but no longer allowed
    CARRY_IN_HANDS = 'carry_in_hands'
    CARRY_ON_BACK = 'carry_on_back'
    CASING_FACILITY = 'casing_facility'
    JOINING_QUEUE = 'joining_queue'
    VEHICLE_MOVING = 'vehicle_moving'


# activity validations are described as regular expressions
# individual actors are encoded as single characters (mapping below)
# words separated by | must be internally sorted in alphabetical order
# because candidates will be sorted before testing to avoid using regex
# lookaround.
#
# for example:
#   `bop` is a valid spec, but `pob` is not.
#   `(kpv|bo)` is valid because each individual word is sorted.
activity_spec = {
    ActivityType.HAND_INTERACTS_WITH_PERSON: 'p{2,}',
    ActivityType.PERSON_ASCENDS_STAIRS: 'p',
    ActivityType.PERSON_ABANDONS_PACKAGE: '(bp|op)',
    ActivityType.PERSON_CARRIES_HEAVY_OBJECT: '(bp+|o+p+|p{2,})',
    ActivityType.PERSON_CLOSES_FACILITY_DOOR: 'p',
    ActivityType.PERSON_CLOSES_TRUNK: 'p?v',
    ActivityType.PERSON_CLOSES_VEHICLE_DOOR: 'p?v',
    ActivityType.PERSON_DESCENDS_STAIRS: 'p',
    ActivityType.PERSON_DISCARDS_TRASH: 'b*o*p',
    ActivityType.PERSON_DONS_JACKET: 'p',
    ActivityType.PERSON_DOFFS_JACKET: 'p',
    ActivityType.PERSON_DRINKS: 'o*p',
    ActivityType.PERSON_EMBRACES_PERSON: 'p{2,}',
    ActivityType.PERSON_ENTERS_SCENE_THROUGH_STRUCTURE: 'p',
    ActivityType.PERSON_ENTERS_VEHICLE: 'pv',
    ActivityType.PERSON_EXITS_SCENE_THROUGH_STRUCTURE: 'p',
    ActivityType.PERSON_EXITS_VEHICLE: 'pv',
    ActivityType.PERSON_INTERACTS_WITH_LAPTOP: 'p',
    ActivityType.PERSON_LEADS_ANIMAL: 'o*p',
    ActivityType.PERSON_LOADS_VEHICLE: '(o*pv|b*pv)',
    ActivityType.PERSON_OPENS_FACILITY_DOOR: 'p',
    ActivityType.PERSON_OPENS_TRUNK: 'p?v',
    ActivityType.PERSON_OPENS_VEHICLE_DOOR: 'pv',
    ActivityType.PERSON_PICKS_UP_OBJECT: '(bp|op+)',
    ActivityType.PERSON_PULLS_OBJECT: '(bp|op+)',
    ActivityType.PERSON_PUSHES_OBJECT: '(bp|op+)',
    ActivityType.PERSON_PURCHASES: 'p+',
    ActivityType.PERSON_READS_DOCUMENT: 'p',
    ActivityType.PERSON_RIDES_BICYCLE: 'kp',
    ActivityType.PERSON_RUMMAGES_IN_PARCEL: '(op|bp)',
    ActivityType.PERSON_SITS_DOWN: 'p',
    ActivityType.PERSON_STANDS_UP: 'p',
    ActivityType.PERSON_STEALS_OBJECT: '(bo*p|o+p)',
    ActivityType.PERSON_TALKS_ON_PHONE: 'p',
    ActivityType.PERSON_TALKS_TO_PERSON: 'p{2,}',
    ActivityType.PERSON_TEXTS_ON_PHONE: 'p',
    ActivityType.PERSON_TRANSFERS_OBJECT: '(o*p{2}|b*p{2}|o*pv|b*pv)',
    ActivityType.PERSON_UNLOADS_VEHICLE: '(o*pv|b*pv)',
    ActivityType.VEHICLE_DROPS_OFF_PERSON: 'p+v',
    ActivityType.VEHICLE_MAKES_U_TURN: 'v',
    ActivityType.VEHICLE_PICKS_UP_PERSON: 'p+v',
    ActivityType.VEHICLE_REVERSES: 'v',
    ActivityType.VEHICLE_STARTS: 'v',
    ActivityType.VEHICLE_STOPS: 'v',
    ActivityType.VEHICLE_TURNS_LEFT: 'v',
    ActivityType.VEHICLE_TURNS_RIGHT: 'v',
}

all_activity_codes = {
    3: ActivityType.PERSON_STANDS_UP,
    4: ActivityType.PERSON_SITS_DOWN,
    7: ActivityType.PERSON_TALKS_ON_PHONE,
    8: ActivityType.PERSON_READS_DOCUMENT,
    9: ActivityType.PERSON_TEXTS_ON_PHONE,
    10: ActivityType.VEHICLE_STOPS,
    11: ActivityType.VEHICLE_STARTS,
    12: ActivityType.VEHICLE_TURNS_LEFT,
    13: ActivityType.VEHICLE_TURNS_RIGHT,
    14: ActivityType.VEHICLE_MAKES_U_TURN,
    15: ActivityType.PERSON_TRANSFERS_OBJECT,
    16: ActivityType.PERSON_PICKS_UP_OBJECT,
    17: ActivityType.PERSON_PUTS_DOWN_OBJECT,
    18: ActivityType.PERSON_CARRIES_HEAVY_OBJECT,
    19: ActivityType.HAND_INTERACTS_WITH_PERSON,
    20: ActivityType.PERSON_EMBRACES_PERSON,
    21: ActivityType.PERSON_OPENS_FACILITY_DOOR,
    22: ActivityType.PERSON_CLOSES_FACILITY_DOOR,
    23: ActivityType.PERSON_ENTERS_SCENE_THROUGH_STRUCTURE,
    24: ActivityType.PERSON_EXITS_SCENE_THROUGH_STRUCTURE,
    25: ActivityType.PERSON_OPENS_VEHICLE_DOOR,
    26: ActivityType.PERSON_CLOSES_VEHICLE_DOOR,
    27: ActivityType.PERSON_ENTERS_VEHICLE,
    28: ActivityType.PERSON_EXITS_VEHICLE,
    29: ActivityType.PERSON_OPENS_TRUNK,
    30: ActivityType.PERSON_CLOSES_TRUNK,
    31: ActivityType.PERSON_LOADS_VEHICLE,
    32: ActivityType.PERSON_UNLOADS_VEHICLE,
    33: ActivityType.VEHICLE_PICKS_UP_PERSON,
    34: ActivityType.VEHICLE_DROPS_OFF_PERSON,
    35: ActivityType.PERSON_TALKS_TO_PERSON,
    37: ActivityType.PERSON_RIDES_BICYCLE,
    38: ActivityType.PERSON_PURCHASES,
    39: ActivityType.PERSON_INTERACTS_WITH_LAPTOP,
    40: ActivityType.PERSON_ABANDONS_PACKAGE,
    42: ActivityType.PERSON_LEADS_ANIMAL,
    43: ActivityType.PERSON_STEALS_OBJECT,
    44: ActivityType.VEHICLE_REVERSES,
    45: ActivityType.PERSON_ASCENDS_STAIRS,
    46: ActivityType.PERSON_DESCENDS_STAIRS,
    47: ActivityType.PERSON_PUSHES_OBJECT,
    48: ActivityType.PERSON_PULLS_OBJECT,
    49: ActivityType.PERSON_DONS_JACKET,
    50: ActivityType.PERSON_DOFFS_JACKET,
    51: ActivityType.PERSON_RUMMAGES_IN_PARCEL,
    52: ActivityType.PERSON_DISCARDS_TRASH,
    53: ActivityType.PERSON_DRINKS,
}


class ActivityPipelineStatuses(StrEnum):
    UNAUDITED = 'unaudited'
    AUDITED = 'audited'
    GLADIATOR_1 = 'gladiator_1'
    REFINER_1 = 'refiner_1'
    GLADIATOR_2 = 'gladiator_2'
    REFINER_2 = 'refiner_2'
    GLADIATOR_3 = 'gladiator_3'
    JANITOR = 'janitor'
    GOOD = 'good'


class AnnotationVendors(StrEnum):
    IMERIT = 'imerit'
    KITWARE = 'kitware'


class CameraLocation(StrEnum):
    ADMIN = 'admin'
    BUS = 'bus'
    SCHOOL = 'school'
    HOSPITAL = 'hospital'


class DataCollects(StrEnum):
    M1 = 'm1'
    M2 = 'm2'


class ReleaseBatches(StrEnum):
    SEQUESTERED = 'sequestered'
    MEVA = 'meva'
    TESTING = 'testing'


class CameraTypes(StrEnum):
    VISIBLE = 'eo'
    INFRARED = 'ir'


class VideoPipelineStatuses(StrEnum):
    ANNOTATION = 'annotation'
    AUDIT = 'audit'
    GUNRUNNER = 'gunrunner'


class Scenarios(StrEnum):
    BASKETBALL = 'basketball'
    CONFERENCE = 'conference'
    CONTROL = 'control'
    DEBATE_TEAM = 'debate team'
    DENSE_SUBWAY = 'dense subway'
    FOOTRACE = 'footrace'
    PANEL_DEBATE = 'panel debate'
    POLITICAL_DEBATE = 'political debate'
    POLITICAL_RALLY = 'political rally'
    THREAT = 'threat'
    TRAVEL = 'travel'



from typing import cast, Dict

from boiler import fakes
from boiler.definitions import ActorType
from boiler.models import Actor
from boiler.serialization import kw18


def test_serialize_types():
    actors = cast(
        Dict[int, Actor],
        {
            1: fakes.ActorFactory(actor_type=ActorType.VEHICLE, begin=0, end=10, detections=[]),
            2: fakes.ActorFactory(actor_type=ActorType.PERSON, begin=5, end=15, detections=[]),
            3: fakes.ActorFactory(actor_type=ActorType.PERSON, begin=2, end=9, detections=[]),
        },
    )
    string_gen = kw18.serialize_types(actors)
    produced = [s for s in string_gen]
    expected = ['1 vehicle\n', '2 person\n', '3 person\n']
    assert produced == expected


def test_serialize_geom():
    b1 = fakes.BoxFactory(left=0, top=1, right=2, bottom=3)
    b2 = fakes.BoxFactory(left=2, top=4, right=6, bottom=8)
    b3 = fakes.BoxFactory(left=5, top=4, right=7, bottom=8)
    d1 = fakes.DetectionFactory(frame=0, box=b1, keyframe=True)
    d2 = fakes.DetectionFactory(frame=1, box=b2, keyframe=False)
    d3 = fakes.DetectionFactory(frame=2, box=b3, keyframe=True)
    actor1 = fakes.ActorFactory(
        actor_type=ActorType.VEHICLE, detections=[d1, d2, d3], begin=0, end=2
    )
    actor2 = fakes.ActorFactory(actor_type=ActorType.PERSON, detections=[d1, d3], begin=0, end=0)
    string_gen = kw18.serialize_geom({1: actor1, 2: actor2}, '=')
    produced = [s for s in string_gen]
    expected = [
        '1 2 0 = = = = = = 0 1 2 3\n',
        '1 2 2 = = = = = = 5 4 7 8\n',
        '2 0 0 = = = = = = 0 1 2 3\n',
        '2 0 2 = = = = = = 5 4 7 8\n',
    ]
    assert produced == expected
